
import shutil
import tempfile
from pathlib import Path, PosixPath
from abc import ABC, abstractmethod
import zipfile
import xmlschema
import os
from lxml import etree as ET

from pyssp_standard.common_content_ssc import Annotations, Annotation, BaseElement, TopLevelMetaData
from pyssp_standard.standard import ModelicaStandard


def register_namespaces():
    for name, url in ModelicaStandard.namespaces.items():
        ET.register_namespace(name, url)


class XMLFile(ABC):
    """
    Base for all xml files
    """

    @abstractmethod
    def __read__(self):
        """
        Fill class with information from xml file
        """
        pass

    @abstractmethod
    def __write__(self):
        """
        Write class data to xml objec
        """
        pass

    @property
    def BaseElement(self):
        return self.base_element

    @property
    def TopLevelMetaData(self):
        return self.top_level_metadata

    def __init__(self, file_path, mode='r'):
        """
        :param mode: [r]ead, [a]ppend or [w]rite
        """
        self.__mode = mode
        if type(file_path) is not PosixPath:
            file_path = Path(file_path)
    
        self.__file_path = file_path
        self.root = None
        self.base_element: BaseElement = BaseElement()
        self.top_level_metadata: TopLevelMetaData = TopLevelMetaData()

        if mode == 'r' or mode == 'a':
            self.__read__()

    def write_to_file(self, filepath):
        xml_string = ET.tostring(self.root, pretty_print=True, encoding='utf-8', xml_declaration=True)
        with open(filepath, 'wb') as file:
            file.write(xml_string)

    def check_compliance(self, schema, namespaces):
        # check name for indications that xsd 1.1 should be used, 1.0 is the default
        if "11" in str(schema):
            schema = xmlschema.XMLSchema11(schema)
        else:
            schema = xmlschema.XMLSchema10(schema)

        if self.__mode in ['a', 'w']:  # Temporary file creation
            with tempfile.TemporaryDirectory(suffix="_pyssp") as temp_dir:
                temp_file_path = Path(temp_dir) / 'tmp.xml'
                self.__write__()
                self.write_to_file(temp_file_path)

                xmlschema.validate(temp_file_path, schema, namespaces=namespaces)
        else:
            xmlschema.validate(self.file_path, schema, namespaces=namespaces)

    @property
    def file_path(self):
        return self.__file_path

    def __save__(self):
        """
        Write xml object to file
        """
        self.write_to_file(self.__file_path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__mode in ['w', 'a']:
            self.__write__()
            self.__save__()


class ModelicaXMLFile(XMLFile, ModelicaStandard):
    """
    Base Class for all spp xml files
    """

    def __read__(self):
        pass

    def __write__(self):
        pass

    def __check_compliance__(self):
        super().check_compliance(self.schemas[self.identifier], self.namespaces)

    @property
    def identifier(self):
        return self.__identifier

    @property
    def annotations(self):
        return self.__annotations

    def add_annotation(self, annotation: Annotation):
        self.annotations.add_annotation(annotation)

    def __enter__(self):
        register_namespaces()
        return self

    def __init__(self, file_path, mode='r', identifier='unknown'):

        self.__identifier = identifier
        self.__annotations = Annotations()
        super().__init__(file_path, mode)


class ZIPFile:
    """
    All operations need to be in context and will be applied against temp dir
    
    """

    def __enter__(self):
        self.__temp_path = Path(tempfile.mkdtemp(prefix="pyssp_"))
        self.__unpacked_path = self.__temp_path / self.file_path.stem

        with zipfile.ZipFile(self.file_path, 'r') as zip_ref:
            zip_ref.extractall(self.__unpacked_path)

        self.__in_context = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.readonly and self.__changed:

            zip_file_path = shutil.make_archive(self.__unpacked_path, 'zip', self.__unpacked_path)
            shutil.copy(zip_file_path, self.save_path)

        shutil.rmtree(self.__temp_path)
        
        self.__in_context = False

    def __init__(self, source_path : Path, target_path: Path=None, readonly=False):
        """
        if target_path is not specified it will overwrite the opened file at exit
        this can be probibited by specifying readonly
        """
        if type(source_path) is not PosixPath:
            source_path = Path(source_path)
    
        self.file_path = source_path
        self.save_path = target_path if target_path != None else source_path
        self.readonly = readonly

        self.__changed = False
        self.__in_context = False
        self.__temp_path = ""
        self.__unpacked_path = ""

    def check_context(self):
        if not self.__in_context:
            raise Exception("Function or variable not accessable unless opened using 'with'")

    @staticmethod
    def get_files(dir : Path):
        return {os.path.relpath(p, dir):p for p in dir.rglob('*')}

    @property
    def unpacked_path(self):
        self.check_context()
        return self.__unpacked_path

    @property
    def __files(self):
        """
        get at dict with [rel_path:abs_path]
        """
        return self.get_files(self.__unpacked_path)

    @property
    def files_rel(self):
        """
        get a list of files relative to zip file
        """
        self.check_context()
        return self.__files.keys()
    
    @property
    def files_abs(self):
        """
        get a list of files, with absolute path within the temp dir
        """
        self.check_context()
        return self.__files.values()

    def get_file_temp_path(self, rel_path):
        """
        translate from rel_path to abs_path
        
        """
        self.check_context()
        return self.__files[rel_path]
        
    def add_file(self, file : Path, rel_path = ""):
        """
        Add something to the resource folder of the ssp.
        :param file: filepath of the object to add.
        :param rel_path: relative path inside archive, omitt '/' in the beginning 
        """
        self.check_context()
        if self.readonly:
            raise Exception("Changes are not allowed in readonly archive")

        self.__changed = True
        rel_path = os.path.join(rel_path, os.path.basename(file))
    
        if rel_path not in self.files_rel:
            shutil.copy(file, self.__unpacked_path / rel_path)

    def remove_file(self, rel_path):
        """
        relative path should be within the file, no slash in the begining
        e.g. SystemStructure.ssd
        """
        self.check_context()
        if self.readonly:
            raise Exception("Changes are not allowed in readonly archive")

        self.__changed = True
        file : Path = self.__unpacked_path / rel_path

        if file.exists():
            file.unlink()
        else:
            raise FileNotFoundError(f"Not found {file}")
        

class SSPElement(ABC):

    @abstractmethod
    def to_element(self) -> ET.Element:
        pass

    @abstractmethod
    def from_element(self, element: ET.Element):
        pass


class EmptyElement(ET.ElementBase):

    def __init__(self, tag, attrib=None, **kwargs):
        super().__init__(tag, attrib, **kwargs)

    def __repr__(self):
        return ET.tostring(self, encoding='utf-8')



from dataclasses import dataclass, asdict
from lxml import etree as ET
from lxml.etree import QName

from pyssp_standard.utils import SSPElement
from pyssp_standard.standard import ModelicaStandard


@dataclass
class BaseUnit:
    kg: int = None
    m: int = None
    s: int = None
    A: int = None
    K: int = None
    mol: int = None
    cd: int = None
    rad: int = None
    factor: float = None
    offset: float = None

    def __init__(self, base_unit: dict):
        for field_name, field_type in self.__annotations__.items():
            value = base_unit.get(field_name)
            if value is not None:
                if not isinstance(value, field_type):
                    try:
                        value = field_type(value)
                    except (TypeError, ValueError):
                        raise ValueError(f"Invalid value type for {field_name}. Expected {field_type}.")
                setattr(self, field_name, value)

    def to_dict(self):
        return {k: str(v) for k, v in asdict(self).items() if v is not None}


class Unit(SSPElement, ModelicaStandard):

    def __init__(self, unit, base_unit: BaseUnit = None):

        self.__root = None
        self.__name = None
        self.__base_unit = None

        if type(unit) is ET.Element:
            self.from_element(unit)
        else:
            self.__name = unit
            self.__base_unit = base_unit

    def to_element(self):
        unit_entry = ET.Element(QName(self.namespaces['ssc'], 'Unit'), attrib={'name': self.__name})
        unit_entry.append(ET.Element(QName(self.namespaces['ssc'], 'BaseUnit'), attrib=self.__base_unit.to_dict()))
        return unit_entry

    def from_element(self, element):
        self.__name = element.attrib.get('name')
        base_unit = element.findall('BaseUnit')[0]
        self.__base_unit = BaseUnit(base_unit.attrib)


class Units(ModelicaStandard):

    def __init__(self, element: ET.Element = None):
        self.__units = []
        self.__root = None

        if element is not None:
            units = element.findall('Unit', self.namespaces)
            for unit in units:
                self.__units.append(Unit(unit))

    def add_unit(self, unit: Unit):
        self.__units.append(unit)

    def element(self, parent_type='ssc'):
        self.__root = ET.Element(QName(self.namespaces[parent_type], 'Units'))
        for unit in self.__units:
            self.__root.append(unit.to_element())
        return self.__root

    def is_empty(self):
        return True if len(self.__units) == 0 else False



from pyssp_standard.common_content_ssc import Enumerations, Annotations, Annotation
from pyssp_standard.unit import Units
from pyssp_standard.utils import ModelicaXMLFile
from pyssp_standard.standard import ModelicaStandard
from lxml import etree as ET
from lxml.etree import QName


class Connection(ModelicaStandard):

    def __init__(self, element=None, *, start_element=None, start_connector=None, end_element=None, end_connector=None):
        self.__root = None

        self.base_element = None
        self.start_element = start_element  # Optional
        self.start_connector = start_connector
        self.end_element = end_element  # Optional
        self.end_connector = end_connector
        self.suppress_unit_conversion: bool = False

        self.transformation = None
        self.annotations = None
        # Connection geometry not connected

        if element is not None:
            self.__read__(element)

    def __read__(self, element):
        self.start_element = element.get('startElement')
        self.start_connector = element.get('startConnector')
        self.end_element = element.get('endElement')
        self.end_connector = element.get('endConnector')

    def __eq__(self, other):
        if self.start_element == other.start_element and self.start_connector == other.start_connector and \
                self.end_element == other.end_element and self.end_connector == other.end_connector:
            return True
        else:
            return False

    def as_element(self) -> ET.Element:
        self.__root = ET.Element(QName(self.namespaces['ssd'], 'Connection'),
                                 attrib={'startElement': self.start_element,
                                         'startConnector': self.start_connector,
                                         'endElement': self.end_element,
                                         'endConnector': self.end_connector})
        return self.__root

    def as_dict(self):
        return {'source': self.start_element, 'signal': self.start_connector,
                'target': self.end_element, 'receiver': self.end_connector}

    def __rep__(self) -> str:
        return f"""source {self.start_element} - {self.start_connector} -> target {self.end_element} - {self.end_connector}""" 


class Connector(ModelicaStandard):

    def __init__(self, element):
        self.name = ""
        self.kind = ""
        self.value_type = ""

        self.__read__(element)

    def __read__(self, element: ET.Element):
        self.name = element.get('name')
        self.kind = element.get('kind')

    def as_dict(self):
        return {'name': self.name, 'kind': self.kind}

class Component(ModelicaStandard):

    def __init__(self, element):
        self.component_type = None
        self.name = None
        self.source = None
        self.implementation = None
        self.connectors = []
        self.parameter_bindings = None
        self.annotations = None

        self.__read__(element)

    def __read__(self, element):
        self.name = element.get('name')
        connectors = element.findall('ssd:Connectors', namespaces=self.namespaces)
        for connector in connectors[0].findall('ssd:Connector', namespaces=self.namespaces):
            self.connectors.append(Connector(connector))

    def as_dict(self):
        return {'name': self.name, 'connectors': [connector.as_dict() for connector in self.connectors]}


class Element(ModelicaStandard):

    def __init__(self, element):
        self.components = []
        self.__read__(element)

    def __read__(self, element):
        components = element.findall('ssd:Component', namespaces=self.namespaces)
        for component in components:
            self.components.append(Component(component))

    def as_dict(self):
        return [component.as_dict() for component in self.components]


class System(ModelicaStandard):

    def __init__(self, system_element: ET.Element):
        self.name = None
        self.element = None
        self.__connections : list[Connection] = []

        self.connectors : list[Connector] = []
        self.parameter_bindings = []
        self.signal_dictionaries = []

        self.__read__(system_element)

    def __read__(self, element):
        self.name = element.get('name', None)
        elements = element.findall('ssd:Elements', namespaces=self.namespaces)
        if len(elements) > 0:
            self.element = Element(elements[0])

        connections = element.findall('ssd:Connections', namespaces=self.namespaces)
        if len(connections) > 0:
            for connection in connections[0].findall('ssd:Connection', namespaces=self.namespaces):
                self.__connections.append(Connection(connection))

    @property
    def connections(self):
        return self.__connections


class DefaultExperiment(ModelicaStandard):

    def __init__(self, element: ET.Element = None):
        self.start_time = None
        self.end_time = None
        self.annotations: Annotations = Annotations()

        if element is not None:
            self.__read__(element)

    def __read__(self, element):
        self.start_time = element.get('startTime')
        self.end_time = element.get('endTime')

        annotations = element.findall('ssd:Annotations', self.namespaces)
        if len(annotations) > 0:
            for annotation in annotations[0].findall('ssc:Annotation', self.namespaces):
                self.annotations.add_annotation(Annotation(annotation))


class SSD(ModelicaXMLFile):

    def __init__(self, file_path, mode='r'):

        self.name = None
        self.version = None

        self.system = None
        self.default_experiment = None
        self.__enumerations: Enumerations = Enumerations()
        self.__units: Units = Units()

        self.connections_to_add = []
        self.connections_to_remove = []

        if mode not in ['r', 'a']:
            raise Exception('Only read mode and append mode are supported for SSD files')

        super().__init__(file_path=file_path, mode=mode, identifier='ssd')

    def __read__(self):
        tree = ET.parse(str(self.file_path))
        self.root = tree.getroot()

        system = self.root.findall('ssd:System', self.namespaces)[0]
        self.system = System(system)

        default_experiment = self.root.findall('ssd:DefaultExperiment', self.namespaces)
        if len(default_experiment) > 0:
            self.default_experiment = DefaultExperiment(default_experiment[0])

        self.name = self.root.get('name')
        self.version = self.root.get('version')

    def __write__(self):
        tree = ET.parse(str(self.file_path))
        self.root = tree.getroot()
        system = self.root.findall('ssd:System', self.namespaces)
        if system is not None:
            connections_set = system[0].findall('ssd:Connections', self.namespaces)[0]
            for target in self.connections_to_remove:
                matching_elements = connections_set.findall('ssd:Connection', namespaces=self.namespaces)
                for element in matching_elements:
                    if Connection(element) == target:
                        element.getparent().remove(element)
            for add_connection in self.connections_to_add:
                connections_set.append(add_connection.as_element())

    def add_connection(self, connection: Connection):
        if type(connection) is not Connection:
            raise "Only Connection object may be used."
        self.system.connections.append(connection)
        self.connections_to_add.append(connection)

    def remove_connection(self, connection: Connection):
        for idx, item in enumerate(self.connections()):
            if item == connection:
                self.system.connections.pop(idx)
                self.connections_to_remove.append(connection)
                break

    def connections(self):
        return self.system.connections

    def list_connections(self, *, start_connector=None, end_connector=None, start_element=None, end_element=None):
        connections = self.connections()
        matching_connections = []

        def check(value, comparision):
            return True if value is None else value == comparision

        for connection in connections:
            connection : Connection
            sc = [start_connector, connection.start_connector]
            ec = [end_connector, connection.end_connector]
            se = [start_element, connection.start_element]
            ee = [end_element, connection.end_element]

            if check(*sc) and check(*ec) and check(*se) and check(*ee):
                matching_connections.append(connection)

        return matching_connections

    def list_connectors(self, *, kind=None, name=None, parent=None):
        """Returns a list of connectors, filtered by the following optional options
            :param kind: the kind of connector, e.g. input, output or parameter
            :param name: the name of the connector, utilizes 'in' for lookup
            :param parent: the name of the parent component, utilizes 'in' for lookup
        """

        matching_connectors = {}
        component_connectors = self.system.element.as_dict()

        for component in component_connectors:
            if parent is not None and parent not in component['name']:
                continue

            for connector in component['connectors']:
                if kind is not None and kind != connector['kind']:
                    continue
                if name is not None and name not in connector['name']:
                    continue

                if component['name'] not in matching_connectors.keys():
                    matching_connectors[component['name']] = []
                matching_connectors[component['name']].append({'name': connector['name'], 'kind': connector['kind']})

        return matching_connectors



import os
import tempfile
import zipfile
import shutil
from pathlib import Path, PosixPath

from pyssp_standard.ssd import SSD
from pyssp_standard.ssb import SSB
from pyssp_standard.ssv import SSV
from pyssp_standard.ssm import SSM
from pyssp_standard.fmu import FMU
from pyssp_standard.standard import ModelicaStandard
from pyssp_standard.utils import ZIPFile


class SSP(ZIPFile):

    def __enter__(self):
        super().__enter__()
        self.ssp_resource_path = self.unpacked_path / 'resources'

        return self

    def __init__(self, source_path, target_path=None, readonly=False):
        super().__init__(source_path, target_path, readonly)
        self.ssp_resource_path :Path = None

    def __rep__(self) -> str:
        spacing = "\t\t"
        return \
f"""{'_'*100}
SSP:
    Path       {self.file_path}
    Temp_dir:  {self.unpacked_path}
    Resources:
{spacing}{spacing.join([str(r) for r in self.resources])}
{'_'*100}
"""

    @property
    def ssd(self):
        ssd = list(self.unpacked_path.glob('*.ssd'))[0]
        return SSD(ssd)

    @property
    def ssv(self):
        ssv = list(self.ssp_resource_path.glob('*.ssv'))
        return [SSV(ssv) for ssv in ssv]

    @property
    def ssm(self):
        ssm = list(self.ssp_resource_path.glob('*.ssm'))
        return [SSM(file) for file in ssm]

    @property
    def ssb(self):
        ssb = list(self.ssp_resource_path.glob('*.ssb'))
        return [SSB(file) for file in ssb]

    @property
    def fmu(self):
        fmu = list(self.ssp_resource_path.glob('*.fmu'))
        return [FMU(file) for file in fmu]

    @property
    def resources(self):
        """ 
        Returns a list of available resources in the ssp folder /resources
        """
        return [f for f in self.get_files(self.ssp_resource_path).keys()]

    def add_resource(self, file :Path):
        """
        Add something to the resource folder of the ssp.
        :param resource: filepath of the object to add.
        """
        self.add_file(file, "resources")

    def remove_resource(self, resource_name):
        """
        Remove resource
        """
        if type(resource_name) is not str:
            resource_name = resource_name.name

        self.remove_file(f"resources/{resource_name}")
        


import datetime
from typing import TypedDict, List
from lxml import etree as ET
from lxml.etree import QName
from dataclasses import dataclass, asdict, fields

from pyssp_standard.standard import ModelicaStandard


class Annotation(ModelicaStandard):  # TODO needs to read and not just create

    def __init__(self, type_declaration):
        """
        The SSP standard allows for the addition of annotations, when created they must contain at least one annotation.
        An annotation may contain anything, however to ease its use the pyssp_standard provides tools to add text, attributes
        elements and ET.Element.
        :param type_declaration: normalized string
        """
        if type(type_declaration) is str:
            self.root = ET.Element(QName(self.namespaces['ssc'], 'Annotation'), attrib={"type": type_declaration})
        elif type(type_declaration) is ET._Element:
            self.root = type_declaration

    def add_element(self, element: ET.Element):
        self.root.append(element)

    def add_text(self, text: str):
        self.root.text = text

    def add_dict(self, name: str, attributes: dict):
        self.root.append(ET.Element(name, attrib=attributes))


class Annotations(ModelicaStandard):

    def __init__(self):
        self.__count = 0
        self.root = ET.Element(QName(self.namespaces['ssc'], 'Annotations'))

    def add_annotation(self, annotation: Annotation):
        self.__count += 1
        self.root.append(annotation.root)

    def element(self):
        return self.root

    def is_empty(self):
        return True if self.__count == 0 else False


@dataclass
class BaseElement:
    id: str = ""
    description: str = ""

    def __repr__(self):
        return f"ID: {self.id} \nDescription: {self.description}"

    def update(self, attributes: dict):
        for key, value in attributes.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def dict(self):
        return {key: str(value) for key, value in asdict(self).items()}

    def update_root(self, root: ET.Element):
        for key, value in self.dict().items():
            if value != "":
                root.set(key, value)
        return root


@dataclass
class TopLevelMetaData:
    author: str = ""
    fileversion: str = ""
    copyright: str = ""
    license: str = ""
    generationTool: str = "pyssp_standard"
    generationDateAndTime: datetime.datetime = datetime.datetime.now().isoformat()

    def __repr__(self):
        base_txt = ""
        for field in fields(self):
            reformatted_text = field.name.replace('_', ' ')
            reformatted_text = reformatted_text.capitalize()
            base_txt += f"{reformatted_text}: {getattr(self, field.name)}\n"
        return base_txt

    def update(self, attributes: dict):
        for key, value in attributes.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def dict(self):
        return {key: str(value) for key, value in asdict(self).items()}

    def update_root(self, root):
        for key, value in self.dict().items():
            if value != "":
                root.set(key, value)
        return root


class Item(TypedDict):
    name: str
    value: int


class Enumeration(TypedDict):
    base_element: BaseElement
    name: str
    items: List[Item]
    annotations: Annotations


class Enumerations(ModelicaStandard):

    def __init__(self, enumerations: List[Enumeration] = None):
        self.__root = ET.Element(QName(self.namespaces['ssc'], 'Enumerations'))
        if enumerations is not None:
            for enum in enumerations:
                self.__root.append(enum)

    def add_enumeration(self, enum: Enumeration):
        self.__root.append(enum)
